implement Ox;
# o/x. In charge of opening files and executing commands from
# applications using o/mero

include "mods.m";
fsname, seled, debug, findpanel, findseled,
sharedns, scroll, trees, ui, Tree, Edit: import oxedit;
Xcmd, deledit, newedit, msg, fatal, putedit: import oxex;
editcmd: import sam;
drop: import str;

Ox: module
{
	init:	 fn(nil: ref Draw->Context, nil: list of string);
};

omero := "/mnt/ui";
persist := 0;

dump()
{
	for(trl := oxedit->trees; trl != nil; trl = tl trl)
		(hd trl).dump();
}

mkhome(s: string)
{
	home := Tree.new(s);
	home.mk(nil);
	newedit(home, s, 0, 0);
}

mkui(dir: string): chan of list of string
{
	dir = names->rooted(workdir->init(), dir);
	dir = names->cleanname(dir);
	ui = Panel.init("ox");
	if(ui == nil)
		fatal(sprint("o/x: can't create ui: %r"));
	ui.ctl("hold\n");
	mkhome(dir);
	ui.ctl("release\n");
	return ui.eventc();
}

look(tr: ref Tree, ed: ref Edit, what: string)
{
	if(ed != nil)
		ed.lru = now();
	run(tr, ed, "Open " + what);
}

lookfile(tr: ref Tree, dir, what: string)
{
	path := names->rooted(dir, what);
	addr, fname: string;

	(e, nil) := stat(fsname(path));
	err: string;
	if(e < 0){
		err = sprint("%r");
		# file:address?
		(fname, addr) = splitl(what, ":");
		fname = names->rooted(dir, fname);
		addr = drop(addr, ":");
		if(len addr > 0 && addr[len addr -1] == ':')
			addr = addr[:len addr-1];
		(e, nil) = stat(fsname(fname));
		if(e >= 0)
			path = fname;
		else
			addr = nil;
	}
	if(e < 0)
		msg(tr, dir, sprint("%s: %s\n", what, err));
	else {
		ned := newedit(tr, path, 0, 0);
		if(addr != nil && ned != nil)
			editcmd(ned, addr + "\n");
	}
}

done(tr: ref Tree): int
{
	for(edl := tr.eds; edl != nil; edl = tl edl)
		pick edp := hd edl{
		File =>
			if(edp.dirty){
				msg(tr, tr.path, sprint("%s: unsaved edits\n", tr.path));
				return -1;
			}
		}
	tr.close();
	return 0;
}

# These commands go to samcmd.b, sharing code both for the command language and
# for the menus.
run(tr: ref Tree, ed: ref Edit, what: string)
{
	stderr = fildes(2);
	dir: string;
	if(ed != nil){
		dir = ed.dir;
		ed.lru = now();
	} else
		dir = tr.path;
	what = drop(what, " \t\n");
	(nwargs, wargs) := tokenize(what, " \t\n");
	if(nwargs < 1)
		return;
	cmd := hd wargs;
	seled = findseled();
	if(seled != nil){
		seled.lru = now();
		seled.getedits();
	}
	if(debug)
		fprint(stderr, "seled: %s \n", seled.text());

	ui.ctl("hold\n");
	case cmd {
	"" =>
		; # ignore
	"End" =>
		if(!persist)
		do {
			if(trees == nil){
				ui.ctl("release\n");
				kill(pctl(0,nil), "killgrp");
				exit;
			}
		} while(done(hd trees) >= 0);
	"Close" =>
		if(ed == nil){
			done(tr);
			if(len trees == 0)
				mkhome(getenv("home"));
		} else
			deledit(ed);
	"Open" =>
		if(nwargs > 1)
			lookfile(tr, dir, hd tl wargs);
		else {
			if(ed != nil)
			if((e := ed.cleanto(cmd, nil)) != nil)
				msg(tr, dir, sprint("%s: %s\n", ed.path, e));
			else
				if(ed.get() < 0)
					ed.close();
		}
	"Write" =>
		if(ed == nil)
			msg(tr, dir, "no edit at point\n");
		else {
			where := ed.path;
			if(nwargs > 1)
				where = hd tl wargs;
			putedit(ed, where);
		}
	"Keep" =>
		if(ed != nil)
			ed.keep = 1;
		else
			msg(tr, dir, "no edit at point\n");
	"Dup" =>
		scr: string;
		if(nwargs == 2){
			scr = hd tl wargs;
			create(omero + "/" + scr, OREAD, 8r775+DMDIR);
		} else
			scr = nil;
		nt := Tree.new(dir);
		nt.mk(scr);
		newedit(nt, dir, 0, 1);
		if(nwargs == 2)
			msg(tr, dir, scr += " created\n");
	"Cmds" =>
		msg(tr, dir, Xcmd.ftext(0));
	"Scroll" =>
		if(nwargs != 1)
			msg(tr, dir, "usage: Scroll\n");
		else
			scroll = !scroll;
			if(scroll)
				msg(tr, dir, "new panels scroll\n");
			else
				msg(tr, dir, "new panels do not scroll\n");
	"Ctl" =>
		if(nwargs < 2)
			msg(tr, dir, "usage: Ctl ctl\n");
		else if(ed != nil){
			ctl := hd tl wargs;
			for(wargs = tl tl wargs; wargs != nil; wargs = tl wargs)
				ctl += " " + hd wargs;
			ed.body.ctl(ctl  + "\n");
		}
			
	"Edit" or "Rc" or "Sh" =>
		ned := newedit(tr, sprint("[%s]", cmd), 1, 0);
		ned.body.ctl("nousel\n");
	* =>
		# external command; it always refers to seled.
		cmd=drop(cmd, " \t\n");
		ck := cmd[0];
		if(ed != nil)
			case ed.path {
			"[Edit]" =>	ck = 'x';
			"[Rc]" =>	ck = ';';
			"[Sh]" =>	ck = '%';
			}
		file := "";
		if(seled != nil){
			dir = seled.dir;
			file = seled.path;
		}
		if(ck == '!'){
			if(oxedit->fsdir != "")
				ck = ';';
			else
				ck = '%';
			what = what[1:];
		}
		case ck {
		';' =>
			if(cmd[0] == ';')
				what = what[1:];
			if(debug)
				fprint(stderr, "host cmd %s\n", what);
			Xcmd.new(what, file, dir, nil, nil, tr.tid, 1);
		'%' =>
			if(cmd[0] == '%')
				what = what[1:];
			if(debug)
				fprint(stderr, "inferno cmd %s\n", what);
			Xcmd.new(what, file, dir, nil, nil, tr.tid, 0);
		* =>
			if(debug)
				fprint(stderr, "edit cmd %s\n", what);
			editcmd(seled, what + "\n");
		}
	}
	ui.ctl("release\n");

}

edevent(ed: ref Edit, op: string, arg: string)
{
	tr := Tree.find(ed.tid);
	if(debug)
		fprint(stderr, "o/x: editev %s: %s [%s]\n", ed.path, op, arg);
	case op {
	"look" =>
		look(tr, ed, arg);
	"exec" or "apply" =>
		run(tr, ed, arg);
	"close" or "interrupt" =>
		ed.close();
	"clean" or "dirty" =>
		pick edp := ed {
		File =>
			edp.dirty = (op == "dirty");
		}
	}
	if(debug)
		dump();
}

trevent(tr: ref Tree, op: string, arg: string)
{
	if(debug)
		fprint(stderr, "o/x: treeev %s: %s [%s]\n", tr.path, op, arg);
	case op {
	"look" =>	;
		look(tr, nil, arg);
	"exec" or "apply" =>
		run(tr, nil, arg);
	"close" or "interrupt" =>
		tr.close();
	}
}

updatecmdstag(tr: ref Tree)
{
	ui.ctl("hold\n");
	t := "cmds: " + Xcmd.ftext(1);
	fd := open(tr.xtag.path+"/data", OWRITE|OTRUNC);
	if(fd != nil){
		data := array of byte t;
		write(fd, data, len data);
	}
	ui.ctl("release\n");
}

eventproc(evc: chan of list of string, xc: chan of int)
{
	if(debug)
		fprint(stderr, "\necho killgrp >/prog/%d/ctl\n\n", pctl(0, nil));
	for(;;){
		alt {
		ev := <-evc =>
			if(ev == nil)
				break;
			if(len ev < 3)
				error("bad event length");
			id := int hd ev;
			what := hd tl tl ev;
			evarg := "";
			if(len ev == 4)
				evarg = hd tl tl tl ev;
			(tr, ed) := findpanel(id);
			if(tr == nil)
				if(hd tl tl ev != "close")
					tr = hd trees;
			if(tr != nil)
			if(ed != nil)
				edevent(ed, what, evarg);
			else
				trevent(tr, what, evarg);
		tid := <-xc =>
			tr := Tree.find(tid);
			if(tr != nil)
				updatecmdstag(tr);
		}
	}
}

init(nil: ref Draw->Context, argv: list of string)
{
	sys = load Sys Sys->PATH;
	err = load Error Error->PATH;
	err->init(sys);
	dat = checkload(load Oxdat Oxdat->PATH, Oxdat->PATH);
	dat->loadmods(sys, err);
	initmods(dat->mods);
	arg = checkload(load Arg Arg->PATH, Arg->PATH);
	arg->init(argv);
	arg->setusage("o/x [-9dnp] [-o dir] [-l dir]");
	uidir: string;
	fsdir := "";
	while((opt := arg->opt()) != 0) {
		case opt {
		'9' =>
			fsdir  = "/mnt/fs";
		'd' =>
			debug = 1;
		'n' =>
			sharedns = 1;
		'o' =>
			omero = arg->earg();
		'p' =>
			persist = 1;
		'l' =>
			uidir = arg->earg();
		* =>
			arg->usage();
		}
	}
	argv = arg->argv();
	if(len argv > 0)
		arg->usage();
	setenv("omero", omero);
	pctl(NEWPGRP, nil);
	stderr = fildes(2);
	os->init();
	panels->init();
	oxedit->init(dat, fsdir);
	regx->init(dat);
	sam->init(dat);
	samcmd->init(dat);
	samlog->init(dat);
	xc := chan[10] of int;	# see xcmdproc 
	oxex->init(dat, xc);
	tblks->init(sys, str, err, 0);
	oxload->init(dat);
	scroll = 1;
	evc: chan of list of string;
	if(uidir == nil){
		dir := getenv("home");
		if(getenv("OCTOPUS") != nil && fsdir == "")
			dir = "/pc" + dir;
		else if(dir == nil)
			dir = "/";
		evc = mkui(dir);
	} else
		evc = oxload->loadui(uidir);
	spawn eventproc(evc, xc);
}
